#include <proteingeometry/dist.h>
#include "pgc.h"
#include "pgc_binding.h"
#include "pgc_conf.h"
#include "pgc_function_interface_proteingeometry.h"

/* Here are the volume functions... */
void *pgc_interface_volume_read_pdb (void **args) {
  struct pgc_record *result = malloc (sizeof (struct pgc_record));
  struct file_record *internal_result =
    (void *) open_and_read_pdb ((char *) args[0], 0);
  if (internal_result != NULL) {
    result->record_type_name = strdup ("pdb_record");
    result->public_fields = malloc (sizeof (void *) * 1);
    result->public_fields[0] = strdup ((char *) args[0]);
    result->struct_data = (void *) internal_result;
  }
  else {
    result->record_type_name = strdup ("null");
    result->public_fields = NULL;
    result->struct_data = NULL;
  }
  result->reference_count = 1;
  return (void *) result;
}

/* This function gives the default values for VolMaxDist and DistCheck, and 
   uses Richards' Method B.  Perhaps it would be prudent to make another
   function that lets the user input values... */
void *pgc_interface_volume_volume (void **args) {
  int i;
  struct pgc_vector *result = malloc (sizeof (struct pgc_vector));
  struct file_records *convenience = 
    ((struct pgc_record *) args[0])->struct_data;
  if (strcmp (((struct pgc_record *) args[0])->record_type_name, 
	      "pdb_record") == 0) {
    result->vector_length = 0;
    result->vector_type = PGC_VECTOR;
    result->vector = NULL;

    NN_Calculate_Volumes (convenience, 2, 64.0, 8.5);
    /* So... we assume here that the struct data has been set up with some
       type of volume data... it may be that this data is incorrect, but that's
       not really our problem. */
    /* We're going to group the resulting volume data into a vector of
       vectors -- a subvector for each residue. */
    result->vector_length = convenience->residue_count;
    result->vector = malloc (sizeof (struct pgc_vector *) * 
			     convenience->residue_count);
    for (i = 0; i < convenience->residue_count; i++) {
      ((struct pgc_vector **) result->vector)[i] = 
	malloc (sizeof (struct pgc_vector));
      ((struct pgc_vector **) result->vector)[i]->vector_type = PGC_FLOAT;
      ((struct pgc_vector **) result->vector)[i]->vector_length = 0;
      ((struct pgc_vector **) result->vector)[i]->vector = NULL;
    }
    for (i = 0; i < convenience->atomnum; i++) {
      int residue = convenience->atoms[i]->resnum - 1;
      struct pgc_vector **ptr = (struct pgc_vector **) result->vector;
      ptr[residue]->vector = realloc (ptr[residue]->vector, sizeof (void *) * 
				      (ptr[residue]->vector_length + 1));
      ((void **) ptr[residue]->vector)[ptr[residue]->vector_length] = 
	malloc (sizeof (double));
      *(((double **) ptr[residue]->vector)[ptr[residue]->vector_length]) = 
	convenience->atoms[i]->volume;
      ptr[residue]->vector_length += 1;
    }
  }
  else {
    fprintf (stderr, "volume: record must be of type \"pdb_record\"\n");
    result->vector_type = PGC_NIL;
    result->vector_length = 0;
    result->vector = NULL;
  }
  return (void *) result;
}

/* Again, here, we should either increase the required number of arguments or
   make another function that gives the user more control... */
void *pgc_interface_volume_surface (void **args) {
  int i;
  struct pgc_vector *result = malloc (sizeof (struct pgc_vector));
  struct file_records *convenience = 
    ((struct pgc_record *) args[0])->struct_data;
  if (strcmp (((struct pgc_record *) args[0])->record_type_name, 
	      "pdb_record") == 0) {
    result->vector_length = 0;
    result->vector_type = PGC_VECTOR;
    result->vector = NULL;

    calculate_surface (convenience, 1.4);
    result->vector_length = convenience->residue_count;
    result->vector = malloc (sizeof (struct pgc_vector *) * 
			     convenience->residue_count);
    for (i = 0; i < convenience->residue_count; i++) {
      ((struct pgc_vector **) result->vector)[i] = 
	malloc (sizeof (struct pgc_vector));
      ((struct pgc_vector **) result->vector)[i]->vector_type = PGC_FLOAT;
      ((struct pgc_vector **) result->vector)[i]->vector_length = 0;
      ((struct pgc_vector **) result->vector)[i]->vector = NULL;
    }
    for (i = 0; i < convenience->atomnum; i++) {
      int residue = convenience->atoms[i]->resnum - 1;
      struct pgc_vector **ptr = (struct pgc_vector **) result->vector;
      ptr[residue]->vector = realloc (ptr[residue]->vector, sizeof (void *) * 
				      (ptr[residue]->vector_length + 1));
      ((void **) ptr[residue]->vector)[ptr[residue]->vector_length] = 
	malloc (sizeof (double));
      *(((double **) ptr[residue]->vector)[ptr[residue]->vector_length]) = 
	convenience->atoms[i]->surface;
      ptr[residue]->vector_length += 1;
    }
  }
  else {
    fprintf (stderr, "surface: record must be of type \"pdb_record\"\n");
    result->vector_type = PGC_NIL;
    result->vector_length = 0;
    result->vector = NULL;
  }
  return (void *) result;
}

void *pgc_interface_volume_packing_eff (void **args) {
  int i;
  char *typesfile = pgc_lookup_conf_entry ("types-file");
  char *residuefile = pgc_lookup_conf_entry ("residues-file");
  char *stdvolsfile = pgc_lookup_conf_entry ("stdvols-file");
  void *result = NULL;
  void **newargs = NULL;
  struct pgc_vector *nullresult = malloc (sizeof (struct pgc_vector));
  nullresult->vector_type = PGC_NIL;
  nullresult->vector_length = 0;
  nullresult->vector = NULL;
  if (typesfile == NULL) {
    fprintf 
      (stderr, 
       "packing_eff: Configuration entry for \'types-file\' not found.\n");
    return (void *) nullresult;
  }
  if (residuefile == NULL) {
    fprintf 
      (stderr, 
       "packing_eff: Configuration entry for \'residues-file\' not found.\n");
    return (void *) nullresult;
  }
  if (stdvolsfile == NULL) {
    fprintf 
      (stderr, 
       "packing_eff: Configuration entry for \'stdvols-file\' not found.\n");
    return (void *) nullresult;
  }
  pgc_free_data (PGC_VECTOR, (void *) nullresult);
  newargs = malloc (sizeof (void *) * 4);
  newargs[0] = args[0];
  newargs[1] = strdup (typesfile);
  newargs[2] = strdup (residuefile);
  newargs[3] = strdup (stdvolsfile);
  result = pgc_interface_volume_packing_eff_advanced (newargs);
  for (i = 0; i < 4; i++)
    free (newargs[i]);
  return result;
}

void *pgc_interface_volume_packing_eff_advanced (void **args) {
  struct pgc_vector *result = malloc (sizeof (struct pgc_vector));
  struct file_records *convenience = 
    ((struct pgc_record *) args[0])->struct_data;
  if (strcmp (((struct pgc_record *) args[0])->record_type_name, 
	      "pdb_record") == 0) {
    int i;
    result->vector_length = 0;
    result->vector_type = PGC_VECTOR;
    result->vector = NULL;

    initial_residues (args[1], args[2], 0);
    calculate_surface (convenience, 1.4);
    NN_Calculate_Volumes (convenience, 4, 64, 8.5);
    COPY_ATOM_VECTS_1TO2 
      (convenience->atoms, surface, convenience->atoms, x, 
       convenience->atomnum);
    COPY_ATOM_VECTS_1TO2 
      (convenience->atoms, volume, convenience->atoms, y, 
       convenience->atomnum);
    read_standard_volumes (args[3]);
    RefVToB (convenience);
    COPY_ATOM_VECTS_1TO2 
      (convenience->atoms, b, convenience->atoms, x, convenience->atomnum);
    CovVDWRadiiToOccB (convenience);

    result->vector_length = convenience->residue_count;
    result->vector = malloc (sizeof (struct pgc_vector *) * 
			     convenience->residue_count);
    for (i = 0; i < convenience->residue_count; i++) {
      ((struct pgc_vector **) result->vector)[i] = 
	malloc (sizeof (struct pgc_vector));
      ((struct pgc_vector **) result->vector)[i]->vector_type = PGC_FLOAT;
      ((struct pgc_vector **) result->vector)[i]->vector_length = 0;
      ((struct pgc_vector **) result->vector)[i]->vector = NULL;
    }
    for (i = 0; i < convenience->atomnum; i++) {
      int residue = convenience->atoms[i]->resnum - 1;
      struct pgc_vector **ptr = (struct pgc_vector **) result->vector;
      ptr[residue]->vector = realloc (ptr[residue]->vector, sizeof (void *) * 
				      (ptr[residue]->vector_length + 1));
      ((void **) ptr[residue]->vector)[ptr[residue]->vector_length] = 
	malloc (sizeof (double));
      if ((convenience->atoms[i]->x > 0) && 
	  (convenience->atoms[i]->volume > 0))
	*(((double **) ptr[residue]->vector)[ptr[residue]->
					     vector_length]) = 
	  convenience->atoms[i]->volume / convenience->atoms[i]->x;
      else
	*(((double **) ptr[residue]->vector)[ptr[residue]->
					     vector_length]) = (double) -1;
      ptr[residue]->vector_length += 1;
    }
  }
  else {
    fprintf (stderr, 
	     "packing_eff_advanced: record must be of type \"pdb_record\"\n");
    result->vector_type = PGC_NIL;
    result->vector_length = 0;
    result->vector = NULL;
  }
  return (void *) result;
}

